/*
 *   ORXONOX - the hottest 3D action shooter ever to exist
 *                    > www.orxonox.net <
 *
 *
 *   License notice:
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *   Author:
 *      Gani Aliguzhinov
 *   Co-authors:
 *      ...
 *
 */

#include "WingmanController.h"


namespace orxonox
{

    RegisterClass(WingmanController);
    
    //ActionpointController contains all common functionality of AI Controllers
    WingmanController::WingmanController(Context* context) : ActionpointController(context)
    {
        RegisterObject(WingmanController);
        this->actionTimer_.setTimer(ACTION_INTERVAL, true, createExecutor(createFunctor(&WingmanController::action, this)));
        this->myLeader_ = 0;
        this->bFirstAction_ = true;

    }

    WingmanController::~WingmanController()
    {
        for (size_t i = 0; i < this->actionpoints_.size(); ++i)
        {
            if(this->actionpoints_[i])
                this->actionpoints_[i]->destroy();
        }
        this->parsedActionpoints_.clear();
        this->actionpoints_.clear();
    }
  
    void WingmanController::XMLPort(Element& xmlelement, XMLPort::Mode mode)
    {
        SUPER(WingmanController, XMLPort, xmlelement, mode);

        //XMLPortParam(SectionController, "target_", setTarget, getTarget, xmlelement, mode).defaultValues(100.0f);
    }
    
    //----in tick, move (or look) and shoot----
    void WingmanController::tick(float dt)
    {   
        if (!this->isActive())
            return;
        if (this->myLeader_ && this->myLeader_->getAction() != Action::FIGHT && this->myLeader_->getAction() !=
            Action::FIGHTALL && this->myLeader_->getAction() != Action::ATTACK)
        {
        }   
        
        SUPER(WingmanController, tick, dt);
    }
    
    //----action for hard calculations----
    void WingmanController::action()
    {

        //----If no leader, find one----
        if (!this->myLeader_)
        {
            ActionpointController* newLeader = (findNewLeader());
            this->myLeader_ = newLeader;
            if (this->myLeader_)
            {
                //spread copyOrientation called equally among the division
                if (this->myLeader_->getIdentifier()->getName() == "SectionController")
                    this->actionCounter_ = 1;
                else
                    this->actionCounter_ = 4;
            }
        }
        //----If have leader, he will deal with logic----
        else
        {

        }
        if (!this->myLeader_)
        {
           ActionpointController::action();
        }
        else if (this->myLeader_)
        {
            if (this->myLeader_->bKeepFormation_ || !(this->myLeader_->getAction() == Action::FIGHT || this->myLeader_->getAction() == Action::FIGHTALL
                || this->myLeader_->getAction() == Action::ATTACK))
            {
                ControllableEntity* myEntity = this->getControllableEntity();
                Vector3 myPosition = myEntity->getWorldPosition();
                if (!this->myLeader_)
                {
                    return;
                }
                ControllableEntity* leaderEntity = this->myLeader_->getControllableEntity();
                Quaternion orient = leaderEntity->getWorldOrientation();
                Vector3 leaderPosition = leaderEntity->getWorldPosition();
                Vector3 targetRelativePosition = getFormationPosition();
                if (!this->myLeader_)
                {
                    return;
                }
                Vector3 targetAbsolutePosition = 
                    (leaderPosition + (orient*WorldEntity::FRONT) * (leaderEntity->getVelocity().length()/5)
                     + (orient* (targetRelativePosition)));
                //let ship finish rotating. also don't call copyOrientation to often as it is a slow function.
                if (this->actionCounter_ % 6 == 0 && !this->bHasTargetOrientation_)
                    this->setAction (Action::FLY, targetAbsolutePosition, orient);
                else
                    this->setAction (Action::FLY, targetAbsolutePosition);
                if ((targetAbsolutePosition - myPosition).length() > this->tolerance_ * 1.5f)
                {
                    this->boostControl();
                }
                else
                {
                   this->getControllableEntity()->boost(false);
                }
            }
            else
            {
                switch (this->myLeader_->getAction())
                {
                    case Action::FIGHT:
                    {
                        if (!this->hasTarget())
                        {
                            this->setTarget(this->myLeader_->getTarget());
                        }
                        break;
                    }
                    case Action::FIGHTALL:
                    {
                        if (!this->hasTarget())
                        {
                            this->setTarget(this->myLeader_->getTarget());
                        }
                        break;
                    }
                    case Action::ATTACK:
                    {
                        if (!this->hasTarget())
                        {
                            this->setTarget(this->myLeader_->getTarget());
                        }
                        break;
                    }
                    default:
                    {
                    
                    }
                }
                if (this->hasTarget())
                {
                    //----choose where to go----
                    this->maneuver();
                    //----fire if you can----
                    this->bShooting_ = this->canFire();                
                }
            }
            
            
        }
        this->actionCounter_ += this->actionCounter_ < 100000 ? 1 : -this->actionCounter_ ;
    }
     
   
    Vector3 WingmanController::getFormationPosition ()
    {
        this->setFormationMode( this->myLeader_->getFormationMode() );
        Vector3* targetRelativePosition;
        this->spread_ = this->myLeader_->getSpread();
        if (this->myLeader_->getIdentifier()->getName() == "DivisionController")
        {
            switch (this->formationMode_){
                case FormationMode::WALL:
                {
                    targetRelativePosition = new Vector3 (2*this->spread_, 0, 0 - this->tolerance_);  
                    break;
                }
                case FormationMode::FINGER4: 
                {
                    targetRelativePosition = new Vector3 (2*this->spread_, 0, this->spread_ - this->tolerance_);  
                    break;
                }
                case FormationMode::DIAMOND: 
                {
                    targetRelativePosition = new Vector3 (2*this->spread_, 0, this->spread_ - this->tolerance_);                  
                    break;
                }
            }
        }
        else
        {

            switch (this->formationMode_){
                case FormationMode::WALL:
                {
                    targetRelativePosition = new Vector3 (-2*this->spread_, 0, 0 - this->tolerance_);  
                    break;
                }
                case FormationMode::FINGER4: 
                {
                    targetRelativePosition = new Vector3 (-2*this->spread_, 0, this->spread_ - this->tolerance_);  
                    break;
                }
                case FormationMode::DIAMOND: 
                {
                    targetRelativePosition = new Vector3 (2*this->spread_, -this->spread_, 0 - this->tolerance_);                  
                    break;
                }
            }
        }
        Vector3 result = *targetRelativePosition;
        delete targetRelativePosition;
        return result;
    }
    //----POST: closest leader that is ready to take a new wingman is returned----
    ActionpointController* WingmanController::findNewLeader()
    {

        if (!this->getControllableEntity())
            return 0;

        //----vars for finding the closest leader----
        ActionpointController* closestLeader = 0;
        float minDistance =  std::numeric_limits<float>::infinity();
        Gametype* gt = this->getGametype();

        for (ObjectList<ActionpointController>::iterator it = ObjectList<ActionpointController>::begin(); it; ++it)
        {
            //----0ptr or not a leader or dead?----
            if (!it || 
                (it->getIdentifier()->getName() != "SectionController" && it->getIdentifier()->getName() != "DivisionController") || 
                !(it->getControllableEntity()))
                continue;
            
            //----same team?----
            if ( !CommonController::sameTeam (this->getControllableEntity(), (it)->getControllableEntity(), gt) )
                continue;
            
            //----check distance----
            float distance = CommonController::distance (it->getControllableEntity(), this->getControllableEntity());
            if (distance < minDistance && !(it->hasWingman()))
            {
                closestLeader = *it;
                minDistance = distance;
            }
           
        }
        if (closestLeader)
        {
            //----Racing conditions----
            if (closestLeader->setWingman(orxonox_cast<ActionpointController*>(this)))
            {
                return closestLeader;
            }

        }

        return 0;
    }




}
